/*  main.c -- TopoVista startup  */

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glut.h>

#include "topo.h"


/*  declare and initialize variables from globals.h  */
#define global
#define initial(v) = v
#include "globals.h"



/*  elevation color map, in ascending order */
/*  (also: sea level is dark blue, handled by special case in the code)  */

Color elevMap[] = {
  { 0.5f, 0.5f, 0.2f },	/* brown */
  { 0.3f, 0.7f, 0.3f },	/* green */
  { 0.3f, 0.7f, 0.3f },
  { 0.5f, 0.9f, 0.4f },	/* lighter green */
  { 0.5f, 0.9f, 0.4f },
  { 0.9f, 0.9f, 0.4f },	/* yellow */
  { 0.9f, 0.9f, 0.4f },
  { 0.9f, 0.7f, 0.4f },	/* orange */
  { 0.9f, 0.7f, 0.4f },
  { 0.8f, 0.8f, 0.7f },	/* light grayish tan */
  { 0.8f, 0.8f, 0.7f },
  { 1.0f, 1.0f, 1.0f },	/* white */
  { 1.0f, 1.0f, 1.0f },
};

int elevCount = sizeof(elevMap) / sizeof(elevMap[0]);	/* number of colors */



static char *bsname(char *f);
static int options(int argc, char *argv[]);
static void usage(char *progname);
static void mkscale(Em *em);
static Rgb *readRGB(char *filename, Em *em);

/*  main  */

int main(int argc, char **argv)
{
  Rgb *rgb;
  double starttime=0.0,endtime;
  int opti;
  Path *path, *tmp;

  glutInit(&argc, argv);		/* initialize GLUT system */

  opti = options(argc, argv);		/* process command-line options */

  if (opti >= argc) {			/* if no files then read stdin */
    gem = readEM(NULL);
    gemname = "stdin";
  } else {
    gem = readEM(argv[opti]);		/* read elevation model */
    gemname = bsname(argv[opti]);	/* name of model */
    opti++;
  }

  if( gem == NULL ) {		/* failed to read EM */
    fprintf(stderr, "Can't read elevation model.\n");
    exit(-1);
  }
  mkscale(gem);			/* configure scaling */
  gemXmax = gem->width;
  gemYmax = gem->height;
  fprintf(stderr,
	  "%s size %d x %d, dimensions %d x %d km, elevation %d to %d m.\n",
	  gemname,
	  gemXmax, gemYmax,
	  (int)(xmeters / 1000 + 0.5), (int)(ymeters / 1000 + 0.5),
	  (int)(gem->dzoffset*gem->dz2m),
	  (int)((gem->dzmax+gem->dzoffset)*gem->dz2m));

  /* ======== read color layers and paths ======== */
  nLayers = 0;
  clayer = NULL;
  if (opti < argc) {
    layerName = (char **)malloc((argc-opti)*sizeof(char *));
    clayer = (Rgb **)malloc((argc-opti)*sizeof(Rgb *));
    if( layerName == NULL || clayer == NULL )
      nospace();
  }
  paths = NULL;
  while (opti < argc) {
    rgb = readRGB(argv[opti], gem);
    if (rgb != NULL) {
      clayer[nLayers] = rgb;
      layerName[nLayers] = bsname(argv[opti]);
      ++nLayers;
    } else {	/* assume it's a path file */
      path = readPaths(argv[opti]);
      if( path != NULL ) {
	/* path itself is the head of a linked list of minipaths */
	tmp = path;
	while( tmp->next != NULL ) tmp = tmp->next;
	tmp->next = paths;
	paths = path;
      } else {
	fprintf(stderr, "Cannot read %s\n", argv[opti]);
      }
    }
    opti++;
  }
  colorLayer = nLayers-1;


  if (tflag) {
    starttime = getTime();
  }

  hierarchy = initHier();		/* initialize hierarchy */

  if (tflag) {
    endtime = getTime();
    fprintf(stderr,"Hierarchy built in %f seconds.\n", endtime-starttime);
  }


  /* initialize viewing parameters */
  eye[0] = 0.0;
  eye[1] = 0.0;
  ref[0] = 0.5 * gemXmax;
  ref[1] = 0.5 * gemYmax;
  setEyeRefHeight();

  /* open windows */
  openCtlWin();			/* control window */
  openOvhWin();			/* overhead (2-D) window */
  openPspWin();			/* perspective (3-D) window */

  /* process events */
  glutMainLoop();

  /*NOTREACHED*/
  return 0;
}

static char *bsname(char *f)
{
  char *s, *e, *r;
  int i;

  if ((s = strrchr(f, '/')) == NULL) {
    if ((s = strrchr(f, '\\')) == NULL)
      s = f;
    else
      s += 1;
  }
  else {
    s += 1;
  }

  if ((e = strchr(s,'.')) == NULL)
    e = s+strlen(s);

  r = (char *)malloc((e-s+1)*sizeof(char));
  if( r == NULL ) nospace();
  i = 0;
  while(s<e)
    r[i++] = *s++;
  r[i] = 0;

  return r;
}

/*  options(argc, argv) -- interpret command-line arguments  */

static int options(int argc, char *argv[])
{
  register int i;
  register char *s;

  for (i = 1; i < argc; ++i)
    {
      s = argv[i];
      if (*s != '-') break;
      switch (*++s)
	{
	case 'd':
	  if (*++s)
	    distThresh = atof(s);
	  else if (++i < argc)
	    distThresh = atof(argv[i]);
	  else
	    usage(argv[0]);
	  break;
	case 'e':
	  if (*++s)
	    errThresh = atof(s);
	  else if (++i < argc)
	    errThresh = atof(argv[i]);
	  else
	    usage(argv[0]);
	  break;
	case 't':
	  tflag = 1;
	  break;
	case 'v':
	  if (*++s)
	    vexagger = atof(s);
	  else if (++i < argc)
	    vexagger = atof(argv[i]);
	  else
	    usage(argv[0]);
	  break;

	default:
	  usage(argv[0]);
	}
    }
  return(i);

}



/*  usage(progname) -- issue usage message, and abort  */

static void usage(char *progname)
{
  fprintf(stderr,
	  "usage: %s [-{dev} value] [-t] [DEMfile [PNMfile ...]]\n", progname);
  fprintf(stderr,
	  "  -d <double>  distance threshold (default=%f)\n", IDISTTHRESH);
  fprintf(stderr,
	  "  -e <double>  error threshold (default=%f)\n", IERRTHRESH);
  fprintf(stderr,
	  "  -t           time the hierarchy and surface creation (default=0)\n");
  fprintf(stderr,
	  "  -v <double>  vertical scaling factor (default=%f)\n", IVEXAGGER);
  exit(1);
}



/*  mkscale(em) -- configure scaling  */
/*  sets xmeters, ymeters, m2grid */
/*  high_water, low_water, dz2contour, err2grid  */

static void mkscale(Em *em)
{
  double a, cv, du, dv;

  /* get center latitude, in radians */
  cv = (em->nwUV.v + em->swUV.v + em->neUV.v + em->seUV.v) / 4.;  /* arc-sec */
  a = cv / ARCSEC_PER_RAD;

  /* get dimensions of elevation model, in arc seconds */
  du = (em->neUV.u - em->nwUV.u + em->seUV.u - em->swUV.u) / 2.;
  dv = (em->nwUV.v - em->swUV.v + em->neUV.v - em->seUV.v) / 2.;

  /* get dimensions in meters */
  xmeters = du * METERS_PER_ARCSEC * cos(a);	/* scales by latitude */
  ymeters = dv * METERS_PER_ARCSEC;		/* always the same */

  /* average of x,y meters per cell to convert meters to EM grid coords */
  m2grid = 1. / ((xmeters / em->width + ymeters / em->height) / 2.);

  /* set high_water and low_water in data coordinates */
  high_water = SEA_LEVEL/em->dz2m - em->dzoffset;
  low_water = -SEA_LEVEL/em->dz2m - em->dzoffset;

  /* stored elevation to contour level */
  assert(em->dzmax > 0);
  dz2contour = (double)(elevCount-1) / em->dzmax;

  /* stored error (0...MAXERR) to meters */
  err2grid = (double)em->dzmax*em->dz2m*m2grid / (double)MAXERR;


  /* Elevation model x,y to longitude,latitude */
  /*
    em->xy2ll = createTransform(em->swXY, em->nwXY, em->neXY, em->seXY,
                                em->swUV, em->nwUV, em->neUV, em->seUV);
			      
    em->ll2xy = createTransform(em->swUV, em->nwUV, em->neUV, em->seUV,
                                em->swXY, em->nwXY, em->neXY, em->seXY);
  */

}



static Rgb *readRGB(char *filename, Em *em)
{
  int x,y;
  unsigned char *red, *green, *blue;
  Transform *pnmll2xy;
  UV swXY, nwXY, neXY, seXY;
  UV s, t;
  Pnm *pnm;
  Rgb *rgb;
  char *pnmname;

  pnm = readPNM(filename);			/* read pnm file */
  if (pnm == NULL) return NULL;

  pnmname = bsname(filename);
  fprintf(stderr, "%s size %d x %d.\n",pnmname, pnm->width, pnm->height);

  if (pnm->ll2xy) {
    pnmll2xy = pnm->ll2xy;
  } else {				/* use default trans */
    fprintf(stderr, "Assuming %s extent equals elevation model extent.\n",
	    pnmname);
    swXY.u = 0;
    swXY.v = 0;
    nwXY.u = 0;
    nwXY.v = pnm->height;
    neXY.u = pnm->width;
    neXY.v = pnm->height;
    seXY.u = pnm->width;
    seXY.v = 0;
    pnmll2xy = createTransform(em->swUV, em->nwUV, em->neUV, em->seUV,
			       swXY, nwXY, neXY, seXY);
  }
	
  rgb = initRGB(em->width,em->height);			/* init an rgb layer */
  if (rgb != NULL) {
    red = rgb->red;
    green = rgb->green;
    blue = rgb->blue;
	
    for(y = 0; y < em->height; ++y) {
      for(x = 0; x < em->width; ++x) {
	s.u = x;
	s.v = y;

	transEMxy2lola(em, s, &t);
	transform(pnmll2xy, t, &s);
			
	if (s.u < pnm->width && s.u >= 0 && s.v < pnm->height && s.v >= 0) {
	  *red = PNMRED(pnm, s.u, s.v);
	  *green = PNMGREEN(pnm, s.u, s.v);
	  *blue = PNMBLUE(pnm, s.u, s.v);
	}
	++red;
	++green;
	++blue;
      }
    }

    rgb->legend.size = pnm->legend.size;
    rgb->legend.entry = pnm->legend.entry;
  }

  free(pnmname);
  free(pnmll2xy);
  free(pnm->red);
  free(pnm->green);
  free(pnm->blue);
  free(pnm);

  return rgb;

}
