/*
 * path.c
 * Created by Xiangyu Zhang, to emit a path on the terrain
 * Rewritten by Will Evans to avoid DFS for each segment of path.
 */
#ifdef WIN32
#include <windows.h>
#endif
#include "topo.h"
#include <GL/gl.h>

/* Use vertex reflection to get neighbor vertices from current tri? */
/* #define REFLECT */

static Color pathColor[] = {
  {1.0f, 0.0f, 0.0f },	/* red */
  {1.0f, 1.0f, 0.0f },	/* yellow */
  {0.0f, 1.0f, 0.0f },	/* green */
  {0.0f, 0.0f, 1.0f },	/* blue */
};
/* number of colors */
static int pathCount = sizeof(pathColor) / sizeof(pathColor[0]);
static int npaths = 0;

static double zscale;
static Hier* hp;
static TriIndex tri;
static Coord2d p1, p2, p3, mid;
static double z1, z2, z3;
static double bz, b0;
static Coord2d rodSW, rodSE, rodNE, rodNW;

static void emitVertex(Coord2f p)
{
  glVertex3f(p[0], p[1], (float)(bz + 0.01));
}

static void setZ(void)
{
  z1 = zscale*EMDATA(gem,p1[0],p1[1]);
  z2 = zscale*EMDATA(gem,p2[0],p2[1]);
  z3 = zscale*EMDATA(gem,p3[0],p3[1]);
  /* set denominator for barycentric coordinate calculation */
  b0=(p2[0]-p1[0])*(p3[1]-p1[1])-(p3[0]-p1[0])*(p2[1]-p1[1]);
  /* b0 shouldn't be 0 */
}

static int setBary(Coord2f p)
     /* sets the global bz (the z value of point p) and
      * returns 0 if the point p is outside the current triangle
      * returns 1 if it is inside.
      */
{
  double b1,b2,b3;

  b1=((p2[0]-p[0])*(p3[1]-p[1])-(p3[0]-p[0])*(p2[1]-p[1]))/b0;
  b2=((p3[0]-p[0])*(p1[1]-p[1])-(p1[0]-p[0])*(p3[1]-p[1]))/b0;
  b3=((p1[0]-p[0])*(p2[1]-p[1])-(p2[0]-p[0])*(p1[1]-p[1]))/b0;
  bz = b1*z1 + b2*z2 + b3*z3;
  if( b1 < 0 || b2 < 0 || b3 < 0 ) return 0;
  else return 1;
}

static void ptLoc( Coord2f p )
/* set "tri" to the triangle that contains the point "p" */
{
  if( ONSURFACE(hp,tri) ) {
    setZ();			/* set the z coords of tri */
    setBary(p);			/* set the barycentric coords of p */
    return;
  }
  MIDPOINT(p1, p2, mid);
  /* Inrod insures that points on the split line locate in the legal child */
  if( LEFT_OR_NO_TURN(mid, p3, p) &&
      (hp->info[LCHILD(tri)] & (PARTINROD | ALLINROD)) ) {
    setCoord(p2,p1); setCoord(p1,p3); setCoord(p3,mid);
    tri = LCHILD(tri);
    ptLoc(p);
  } else {
    setCoord(p1,p2); setCoord(p2,p3); setCoord(p3,mid);
    tri = RCHILD(tri);
    ptLoc(p);
  }
}

static int pointLocate( Coord2f p )
{
  if( LEFT_OR_NO_TURN(sw, ne, p) &&
      LEFT_OR_NO_TURN(ne, nw, p) &&
      LEFT_OR_NO_TURN(nw, sw, p) ) {
    setCoord(p1,sw); setCoord(p2,ne); setCoord(p3,nw);
    tri = LCHILD(0);
    ptLoc(p);
    return 1;
  }
  if( LEFT_OR_NO_TURN(ne, sw, p) &&
      LEFT_OR_NO_TURN(sw, se, p) &&
      LEFT_OR_NO_TURN(se, ne, p) ) {
    setCoord(p1,ne); setCoord(p2,sw); setCoord(p3,se);
    tri = RCHILD(0);
    ptLoc(p);
    return 1;
  }
  tri = 0;	/* 0 is not a legal triangle index */
  return 0;
}

#define EPSILON 0.000001f
static int clipLeft( Coord2d b1, Coord2d b2, Coord2f p, Coord2f q )
     /* return 0 if segment p,q is to right of line b1,b2
      * return 1 if segment p,q is to left of  line b1,b2
      * return 2 if p is left of and q is right of line b1,b2 and change q
      * return 3 if q is left of and p is right of line b1,b2 and change p
      */
{
  int dbx = b2[0] - b1[0];
  int dby = b2[1] - b1[1];
  float num, den;
  float t;

  num = (b1[0] - p[0])*dby - (b1[1] - p[1])*dbx;
  den = (q[0] - p[0])*dby - (q[1] - p[1])*dbx;
  if( den == 0 ) {	/* parallel */
    if( num < 0 ) return 0;
    return 1;
  }
  t = num / den;
  if( den > 0 ) {	/* in to out */
    if( t < 1 ) {
      if( t >= 0 ) {
	t -= EPSILON;	/* make sure new q lies inside halfplane */
	if( t < 0 ) t = 0;
	q[0] = p[0] + t*(q[0]-p[0]);
	q[1] = p[1] + t*(q[1]-p[1]);
	return 2;
      }
      return 0;
    }
    return 1;
  }
  /* out to in */
  if( t > 0 ) {
    if( t <= 1 ) {
      t += EPSILON;	/* make sure new p lies inside halfplane */
      if( t > 1 ) t = 1;
      p[0] = p[0] + t*(q[0]-p[0]);
      p[1] = p[1] + t*(q[1]-p[1]);
      return 3;
    }
    return 0;
  }
  return 1;
}

static int clipROD( Coord2f p, Coord2f q )
{
  /* clip to rectangle of definition */
  if( !clipLeft(rodSW, rodSE, p, q) ) return 0;
  if( !clipLeft(rodSE, rodNE, p, q) ) return 0;
  if( !clipLeft(rodNE, rodNW, p, q) ) return 0;
  if( !clipLeft(rodNW, rodSW, p, q) ) return 0;
  return 1;
}

#ifdef REFLECT
static void reflect(Coord2d a, Coord2d b, Coord2d c, Coord2d d)
     /* Reflect a about line b,c to get d.
	Line b,c must be vertical, horizontal, or diagonal.
     */
{
  if( b[0] == c[0] ) {
    d[0] = 2*b[0] - a[0];
    d[1] = a[1];
  } else if( b[1] == c[1] ) {
    d[0] = a[0];
    d[1] = 2*b[1] - a[1];
  } else if( a[0] == b[0] ) {
    d[1] = b[1];
    if( a[1] == c[1] ) {
      d[0] = c[0];
    } else {
      d[0] = 2*c[0] - a[0];
    }
  } else if( a[0] == c[0] ) {
    d[1] = c[1];
    if( a[1] == b[1] ) {
      d[0] = b[0];
    } else {
      d[0] = 2*b[0] - a[0];
    }
  } else if( a[1] == b[1] ) {
    d[0] = b[0];
    d[1] = 2*c[1] - a[1];
  } else if( a[1] == c[1] ) {
    d[0] = c[0];
    d[1] = 2*b[1] - a[1];
  } else {
    fprintf(stderr, "Bad case in reflect({%d,%d},{%d,%d},{%d,%d})\n",
	    a[0],a[1],b[0],b[1],c[0],c[1]);
  }
}
#endif

static TriIndex clipExit(Coord2f p, Coord2f q)
{
  TriIndex nbr = 0;
  Coord2f pp, qq;
#ifdef REFLECT
  Coord2d r;
  int s = 0;
#endif

  pp[0] = p[0]; pp[1] = p[1];
  qq[0] = q[0]; qq[1] = q[1];
  /* Since p is in tri, clipLeft will never return 0. */
  if( clipLeft(p1, p2, pp, qq) == 2 ) {
#ifdef REFLECT
    s = 3;
#endif
    nbr = nbr3(hp,tri);
  }
  if( clipLeft(p2, p3, pp, qq) == 2 ) {
#ifdef REFLECT
    s = 1;
#endif
    nbr = nbr1(hp,tri);
  }
  if( clipLeft(p3, p1, pp, qq) == 2 ) {
#ifdef REFLECT
    s = 2;
#endif
    nbr = nbr2(hp,tri);
  }
  setBary(qq);	/* set bz.  should always return true. */
  emitVertex(qq);
  p[0] = qq[0]; p[1] = qq[1];	/* set p to the exiting vertex */
  if( nbr == 0 || !INROD(hp,nbr) ) return 0;
  /* set the next tri */
#ifdef REFLECT
  /* Reflect one of tri's vertices across the opposite edge to
     get the new nbr vertex.
  */
  switch( s ) {
  case 1:
    reflect(p1,p2,p3,r);
    if( SMALLN1(hp,tri) ) {
      setCoord(p1,p3); MIDPOINT(r,p2,p3);
    } else {
      setCoord(p1,p2); setCoord(p2,r);
    }
    break;
  case 2:
    reflect(p2,p3,p1,r);
    if( SMALLN2(hp,tri) ) {
      setCoord(p2,p3); MIDPOINT(r,p1,p3);
    } else {
      setCoord(p2,p1); setCoord(p1,r);
    }
    break;
  case 3:
    reflect(p3,p1,p2,r);
    if( BIGN3(hp,tri) ) {
      if( tri & 1 ) {	/* left child */
	setCoord(p3,p1); reflect(p2,p1,r,p1);
      } else {		/* right child */
	setCoord(p3,p2); reflect(p1,p2,r,p2);
      }
    } else {
      setCoord(p3,r); setCoord(r,p1); setCoord(p1,p2); setCoord(p2,r);
    }
    break;
  }
#else
  /* Descend from the hierarchy root to get the neighbor vertices.
     This seems to be only slightly less efficient than
     calculating the neighbor vertices from tri using vertex reflection.
  */
  getVerts(nbr, p1, p2, p3);
#endif
  tri = nbr;
  setZ();
  return nbr;
}

static void emitPath(Path* path)
     /* This is the main rountine.
      * 1) Find the point p and triangle tri where the path first
      *    intersects ROD.
      * 2) Emit vertices in tri until the path leaves tri.
      * 3) Use clipExit to find (and emit) the last vertex on tri and
      *    set p to that vertex.
      * 4) If the path hasn't left ROD, set tri to the just entered
      *    neighbor and goto 2, otherwise goto 1.
      */
{
  Coord2f* a;
  Coord2f p, q;
  int n,i;

  a = path->p;
  n = path->n;
  i = 0;
  while( 1 ) {
    while( i+1 < n ) {
      p[0] = a[i][0]; p[1] = a[i][1];
      q[0] = a[i+1][0]; q[1] = a[i+1][1];
      if( clipROD(p, q) ) break;
      ++i;
    }
    ++i;
    if( i >= n ) break;
    pointLocate(p);	/* set current triangle and z-coord of p */
    glBegin(GL_LINE_STRIP);
    /*
      glPointSize(5);
      glBegin(GL_POINTS);
    */
    emitVertex(p);
    while( 1 ) {
      while( i < n && setBary(a[i]) ) {
	emitVertex(a[i]);
	p[0] = a[i][0]; p[1] = a[i][1];
	++i;
      }
      if( i >= n || !clipExit(p, a[i]) ) break;
    }
    glEnd();
  }
}

void emitPaths(Hier* hier, Path* ps)
{
  Path *pth;
  
  hp = hier;
  zscale = vexagger * m2grid * gem->dz2m;
  rodSW[0] = 0; rodSW[1] = 0;
  rodSE[0] = gemXmax-1; rodSE[1] = 0;
  rodNE[0] = gemXmax-1; rodNE[1] = gemYmax-1;
  rodNW[0] = 0; rodNW[1] = gemYmax-1;
  pth = ps;
  while( pth ) {
    glColor3f(pth->color.red, pth->color.green, pth->color.blue);
    emitPath(pth);
    pth = pth->next;
  }
}
      
/* Read file of lon. lat. coords and add to front of "path" linked list */
Path* readPaths(char* filename)
{
  FILETYPE f;
  UV lola,xy;
  Path* tmp;
  Path* path = NULL;
  char buf[512];
  int i, endpath;

  if( (f = FOPEN(filename,"rb")) == NULL ) return NULL;
  endpath = 1;		/* trigger initial path allocation */
  while( FGETS(buf,512,f) ) {
    i = strlen(buf);
    if( i == 0 ) { endpath = 1; continue; }	/* blank lines mean new path */
    i--;
    while( i >= 0 && isspace(buf[i]) ) --i;
    while( i >= 0 && !isspace(buf[i]) ) --i;
    if( sscanf(buf+i+1, "%lf", &(lola.u)) < 1 ) { endpath = 1; continue; }
    while( i >= 0 && isspace(buf[i]) ) --i;
    while( i >= 0 && !isspace(buf[i]) ) --i;
    if( sscanf(buf+i+1, "%lf", &(lola.v)) < 1 ) { endpath = 1; continue; }

    if( endpath ) {
      tmp = (Path *)malloc(sizeof(Path));
      if( tmp == NULL ) nospace();
      tmp->p = NULL;
      tmp->next = path;
      tmp->n = tmp->allocation = 0;
      tmp->color.red = pathColor[npaths].red;
      tmp->color.green = pathColor[npaths].green;
      tmp->color.blue = pathColor[npaths].blue;
      path = tmp;
      endpath = 0;
    }
    lola.u *= 3600.0;
    lola.v *= 3600.0;
    datum2datum(WGS84, lola, gem->datum, &lola);
    transEMlola2xy(gem, lola, &xy);

    if( path->n >= path->allocation ) {
      path->allocation += path->allocation + 1;
      path->p = (Coord2f *)realloc(path->p, path->allocation*sizeof(Coord2f));
      if( path->p == NULL ) nospace();
    }
    path->p[path->n][0] = (float)xy.u;
    path->p[path->n][1] = (float)xy.v;
    path->n++;
  }
  FCLOSE(f);
  npaths = (npaths + 1) % pathCount;
  return path;
}
