/*  view.c -- display perspective view  */

#ifdef WIN32
#include <windows.h>
#endif
#include "topo.h"
#include <GL/gl.h>
#include <GL/glut.h>

static void pspDisplayCB(void);

static int newSurface = 1;	/* does surface need to be recomputed? */

static double angle = 45.0;

#define NEAR_EYE_UP 1
#define NEAR_EYE_DOWN 0.1
static double myfar;
static double aspect;
static GLdouble modelMatrix[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
static GLdouble projMatrix[16];
static GLint viewport[4];

/*  viewChanged() -- take note that view has changed, so surface will, too  */
void viewChanged(void)
{

#ifdef TESTING
  if (doingLoop) {
    newSurface = 1;
    if (doRedraw) {
      glutSetWindow(ovhWin);
      ovhDisplayCB();	/* call it directly so GLUT doesn't take control */
      glutSetWindow(pspWin);
      pspDisplayCB();	/* call it directly so GLUT doesn't take control */
      glutSetWindow(ovhWin);
    } else {
      buildSurface();
    }
    triPerLoop += hierarchy->numSurfTri;
    return;
  }
#endif

  newSurface = 1;
  glutSetWindow(ovhWin);
  glutPostRedisplay();
  glutSetWindow(pspWin);
  glutPostRedisplay();
}


static int proj2plane(GLdouble x, GLdouble y, GLdouble z, Coord2f xy)
{
  double t = -1.0;

  if( z != eye[2] )
    t = eye[2] / (eye[2] - z);
  if( t < 0 ) {
    xy[0] = x; xy[1] = y;
    return 0;
  }
  xy[0] = eye[0] + t * (x - eye[0]);
  xy[1] = eye[1] + t * (y - eye[1]);
  return 1;
}

/* set the projection matrix for the perspective view */
static void setPspProjMatrix( void )
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective((GLdouble)(angle/aspect),(GLdouble)aspect,
		 (GLdouble)(eyeUp ? NEAR_EYE_UP : NEAR_EYE_DOWN),
		 (GLdouble)myfar);

  gluLookAt((GLdouble)eye[0], (GLdouble)eye[1], (GLdouble)eye[2],
	    (GLdouble)ref[0], (GLdouble)ref[1], (GLdouble)ref[2],
	    (GLdouble)0, (GLdouble)0, (GLdouble)1);
}


void setEyeRefHeight(void)
{
  int x, y, z;

  /* compute eye height depending on location and eyeUp flag */
  if (eyeUp) {

    /* eye is up: EYE_HEIGHT meters above highest peak */
    eye[2] = vexagger * m2grid * (gem->dzmax * gem->dz2m + EYE_HEIGHT);

  } else {
    /* eye down: just above max height of gem cell corners */
    x = eye[0];
    y = eye[1];
    z = zval(x, y);
    z = MAX(z, zval(x + 1, y));
    z = MAX(z, zval(x + 1, y + 1));
    z = MAX(z, zval(x, y + 1));
    eye[2] = vexagger * m2grid * (z * gem->dz2m + EYE_HEIGHT);
  }

  /* compute reference height */
  ref[2]=vexagger*m2grid*(zval((int)ref[0],(int)ref[1])*gem->dz2m+REF_HEIGHT);
}

/*  buildSurface() -- rebuild the surface if newSurface is set  */
void buildSurface(void)
{
  GLdouble x, y, z;
  GLdouble origProjMatrix[16];

  if (!newSurface)		/* if nothing's changed, don't rebuild */
    return;

  setEyeRefHeight();		/* set the height of eye and ref points */

  /* Calculate projection matrix for FOV determination */
  /* save original proj matrix */
  glGetDoublev(GL_PROJECTION_MATRIX, origProjMatrix);
  /* set proj matrix to perspective */
  setPspProjMatrix();
  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);
  /* restore saved matrix */
  glMatrixMode(GL_PROJECTION);
  glLoadMatrixd(origProjMatrix);

  gluUnProject(0.0, viewport[3]-1, 1.0,
	       modelMatrix, projMatrix, viewport, &x, &y, &z);
  if( proj2plane(x, y, z, fovULpt) ) {
    /* Ray from eye through upper left corner of view window hits z=0 plane. */
    /* This means FOV is finite, either 4gon or 5gon. */
    gluUnProject(viewport[2]-1, viewport[3]-1, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovURpt);
    gluUnProject(0.0, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLLpt);
    gluUnProject(viewport[2]-1, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLRpt);

    fovTnorm[0] = fovULpt[1] - fovURpt[1];
    fovTnorm[1] = fovURpt[0] - fovULpt[0];
    fovLnorm[0] = fovLLpt[1] - fovULpt[1];
    fovLnorm[1] = fovULpt[0] - fovLLpt[0];
    fovRnorm[0] = fovURpt[1] - fovLRpt[1];
    fovRnorm[1] = fovLRpt[0] - fovURpt[0];
    if( LEFT_TURN(eye, fovLRpt, fovLLpt) ) {
      fovType = FIVE_GON;
      fovBLnorm[0] = eye[1] - fovLLpt[1];
      fovBLnorm[1] = fovLLpt[0] - eye[0];
      fovBRnorm[0] = fovLRpt[1] - eye[1];
      fovBRnorm[1] = eye[0] - fovLRpt[0];
    } else {
      fovType = FOUR_GON;
      fovBLnorm[0] = fovLRpt[1] - fovLLpt[1];
      fovBLnorm[1] = fovLLpt[0] - fovLRpt[0];
    }
  } else {
    fovType = WEDGE;
    gluUnProject(0.0, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLLpt);
    fovLnorm[0] = eye[1] - fovLLpt[1];
    fovLnorm[1] = fovLLpt[0] - eye[0];
    gluUnProject(viewport[2]-1, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLRpt);
    fovRnorm[0] = fovLRpt[1] - eye[1];
    fovRnorm[1] = eye[0] - fovLRpt[0];
  }

  /* build the surface */
  errSurface(hierarchy);

  newSurface = 0;
}

static void pspReshapeCB(int w, int h)
{
  glViewport(0,0,w,h);
  glGetIntegerv(GL_VIEWPORT, viewport);
  aspect = (double)w / (double)h;
}


/*  pspDisplayCB -- redraw perspective (3-D) view  */
static void pspDisplayCB(void)
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  setPspProjMatrix();

  buildSurface();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  renderPsp();
  glutSwapBuffers();
}

/* renderPsp - used by pspDisplayCB and outputEPS */
void renderPsp(void)
{
  if (threeDMesh) {
    emitSurface(hierarchy, WIRE3D);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0,1.0);
  }
  if( smoothSurface ) surfaceNormal(hierarchy);

  emitSurface(hierarchy, FILL3D);
  /*modified by Xiangyu Zhang, for path rendering*/
  /*  glClear(GL_DEPTH_BUFFER_BIT);*/
  emitPaths(hierarchy, paths);

  
  if (threeDMesh) {
    glDisable(GL_POLYGON_OFFSET_FILL);
  }
}

/*  zval(x, y) -- return z value from gem, or 0 if out of bounds  */

int zval(int x, int y)
{
  if (x < 0 || y < 0 || x >= gem->width || y >= gem->height)
    return 0;
  else
    return EMDATA(gem, x, y);
}



/*  openPspWin() -- open perspective view window  */
void openPspWin(void)
{
  GLfloat light_specular[] = { 0.2F, 0.2F, 0.2F, 1.0F };
  GLfloat light_diffuse[] = { 1.0F, 1.0F, 1.0F, 1.0F };
  GLfloat light_ambient[] = { 0.0F, 0.0F, 0.0F, 1.0F };
  GLfloat light_position[] = {0.0F, 1.0F, 2.0F, 0.0F};/*high in northern sky*/
  char wname[200];

  myfar = 8 * MAX(gemXmax,gemYmax);
  aspect = (double)WIDTH_3D / (double)HEIGHT_3D;
  viewport[0] = viewport[1] = 0;
  viewport[2] = WIDTH_3D;
  viewport[3] = HEIGHT_3D;

  glutInitWindowSize(WIDTH_3D, HEIGHT_3D);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  sprintf(wname, "%s", gemname);
  pspWin = glutCreateWindow(wname);

  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  /*  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);*/
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  glCullFace(GL_BACK);		/* enabled/disabled in control.c */
  
  glEnable(GL_NORMALIZE);	/* so OpenGL normalizes normal vectors */

  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);
  glClearColor(0.0,0.0,1.0,0.0);
  /*  glClearColor(0.9,0.9,0.9,0.0);*/
  glClearDepth((GLclampd)myfar);

  glutDisplayFunc(pspDisplayCB);
  glutReshapeFunc(pspReshapeCB);
  glutKeyboardFunc(keyCB);
  glutSpecialFunc(specialCB);
}
